// wuiregis.c - Definitions for WUIman object REGIStration code.
// Copyright (c) 1994 by Ron Burk

// About the only reason this is in a separate module is
// because the code should only be needed at startup and
// shutdown, so you might want to keep it in a separate
// code segment or page.

#include "wuiobjdb.h"

/*

The basic idea here is to implement a single class that provides an
array of Schwartz counters.  When the first static TWuiRegister object
is constructed for a given WUIMAN class, it calls the corresponding
ClassNew() function to create the master object for that class, and adds
it to a list, along with a count of 1.  Any further static TWuiRegister
objects that register this class merely increment the count.  When these
static objects get destroyed at shutdown, the corresponding count is
decremented.  When the count reaches zero, the master object's
ClassDelete() function is called, it is removed from the list, and then
deleted. 

 */

class TWuiMaster
    {
public:
    TWuiMaster(const TWuiName NewName, WUI_FINIFUNC D);
    const TWuiName  Name;
    int             Count;
    WUI_FINIFUNC    Destructor;
    };

TWuiMaster::TWuiMaster(const TWuiName NewName, WUI_FINIFUNC D)
        : Name(NewName), Count(1), Destructor(D)
        {}

static  TWuiGenericList *MasterObjects;

static TWuiMaster *GetItem(const TWuiName Name)
    {
    ASSERTDATAPTR(MasterObjects);
    TWuiMaster *Item = 0;
    for(int i=0; i < MasterObjects->NElements(); ++i)
        {
        ASSERT(i < 256);    // ??? sanity check ???
        Item = (TWuiMaster *)MasterObjects->Get(i);
        ASSERTDATAPTR(Item);
        if(Item->Name == Name)
            return Item;
        }
    return NULL;
    }

TWuiRegister::TWuiRegister(const TWuiName Name,
            WUI_INITFUNC ClassNew, WUI_FINIFUNC ClassDelete)
    {   MEMBERASSERT();
#if 0
#ifndef NDEBUG
    char    Buffer[TWuiName::MAXNAME+1];
    Name.GetName(Buffer);
    DEBUG_TRACE("TWuiRegister('%s')", Buffer);
#endif
#endif

    // initialize list of master objects if necessary
    if(MasterObjects == NULL)
        {
        MasterObjects   = new TWuiGenericList(32);
        ASSERTDATAPTR(MasterObjects);
        }
    TWuiMaster *Item = GetItem(Name);
    // if this class is already registered
    if(Item != NULL)
        {
        ASSERTDATAPTR(Item);
        // then just increment its usage count
        ++Item->Count;
        }
    else
        // else, add it to the list and call its ClassNew()
        {
        AWuiObject *NewMaster = ClassNew(Name);
        Item    = new TWuiMaster(Name, ClassDelete);
        ASSERT(Item != NULL);
        MasterObjects->Insert((void *)Item);
        if(NewMaster != NULL)
            {
            ASSERTDATAPTR(NewMaster);
            RegisterMaster(NewMaster);
            }
        }
    Master  = Item; // store for later destruction
    }

TWuiRegister::~TWuiRegister()
    {   MEMBERASSERT();
    // first, locate this object on the list
    ASSERTDATAPTR(Master);
    // if not last usage
    if(--Master->Count > 0)
        // then just decrement usage count
        return;
    else
        // else, give class chance to de-initialize
        {
        if(Master->Destructor != NULL)
            Master->Destructor(Master->Name);
        delete MasterObjects->Remove((void *)Master);
        }
    }

// TWuiRegister - Register a new WUIMAN class.
void TWuiRegister::RegisterMaster(AWuiObject *MasterCopy)
    {   MEMBERASSERT();
    if(AWuiObject::TraceFlag)
        {
        char Name[TWuiName::MAXNAME];
        MasterCopy->Name().GetName(Name);
        DEBUG_TRACE("TWuiRegister(\"%s\")", Name);
        }

    AWuiObject  *Classes = 0;
    if(AWuiObject::Root == NULL)    // if not initialized
        {
        // create root container object
        AWuiObject::Root    = new TWuiRoot(AWuiObject::S_ROOT);
        ASSERT(AWuiObject::Root != NULL);
        // add "/Classes", to hold registered classes
        Classes             = new TWuiContainer(AWuiObject::S_CLASSES);
        ASSERT(Classes != NULL);
        ASSERT(AWuiObject::Root->Children() != NULL);
        Classes->SetFlags(AWuiObject::PERSISTENT, 0);
        AWuiObject::Root->Children()->Insert(Classes);
        // Initialize 'MasterObjects', master list
        ASSERT(Classes->Children() != NULL);
        AWuiObject::MasterObjects = Classes->Children();
        }
    ASSERT(AWuiObject::MasterObjects != NULL);
    // Place object of class "X" in "/Classes/X".
    int Position=(MasterCopy->IsA(AWuiObject::C_TATTRIBUTE) ? 0 : -1);
    AWuiObject::MasterObjects->Insert(MasterCopy,Position);
    }



